/// This is basically what Bevy master currently does
/// Changing UsizeRes to UsizeRes<'static> makes "unsafety" possible because the param can be moved outside of the call

trait SystemParam<'a> {
    unsafe fn get_param(world: &World) -> Self;
}

struct UsizeRes<'a> {
    value: &'a usize,
}

impl<'a> SystemParam<'a> for UsizeRes<'a> {
    unsafe fn get_param(world: &World) -> Self {
        UsizeRes {
            value: std::mem::transmute(&world.value1),
        }
    }
}

trait IntoSystem<'a, Params, Sys: System> {
    fn system(self) -> Sys;
}

trait System: 'static {
    fn run(&mut self, world: &World);
}

struct FuncSystem {
    sys: Box<dyn FnMut(&World)>,
}

impl System for FuncSystem {
    fn run(&mut self, world: &World) {
        (self.sys)(world);
    }
}

impl<'a, P: SystemParam<'a>, P1: SystemParam<'a>, F: FnMut(P, P1) + 'static>
    IntoSystem<'a, (P, P1), FuncSystem> for F
{
    fn system(mut self) -> FuncSystem {
        FuncSystem {
            sys: Box::new(move |world: &World| unsafe {
                let param = P::get_param(world);
                let param1 = P1::get_param(world);
                self(param, param1);
            }),
        }
    }
}

struct World {
    value1: usize,
}

struct App {
    world: World,
    system: Option<Box<dyn System>>,
}

impl App {
    fn add_system<'a, Params, S: System, IntoS: IntoSystem<'a, Params, S>>(
        &'a mut self,
        into: IntoS,
    ) {
        let system = into.system();
        self.system = Some(Box::new(system));
    }

    fn run(&mut self) {
        self.system.as_mut().unwrap().run(&self.world);
    }
}

fn main() {
    fn sys(res: UsizeRes, res2: UsizeRes) {
        println!("{}", res.value);
    }

    let world = World { value1: 10 };
    // let mut system = sys.system();
    // system.run(&world);

    let mut app = App {
        world,
        system: None,
    };

    app.add_system(sys);
    app.run();
}
