/// This is the best option ive found so far
/// It is basically a combination of Bevy 0.3 and Bevy master techniques
/// It is safe: changing UsizeRes to UsizeRes<'static> causes a compile time error
/// Unfortunately into_system doesn't work for raw systems (aka sys.system() is required)

struct UsizeRes<'a> {
    value: &'a usize,
}

trait SystemParam {
    type Fetch: for<'a> FetchSystemParam<'a>;
}

trait FetchSystemParam<'a> {
    type Item;
    fn get_param(world: &'a World) -> Self::Item;
}

impl<'a> SystemParam for UsizeRes<'a> {
    type Fetch = FetchUsizeRes;
}

struct FetchUsizeRes;

impl<'a> FetchSystemParam<'a> for FetchUsizeRes {
    type Item = UsizeRes<'a>;
    fn get_param(world: &'a World) -> Self::Item {
        UsizeRes {
            value: &world.value1,
        }
    }
}

trait IntoSystem<Params, Sys: System> {
    fn system(self) -> Sys;
}

trait System: 'static {
    unsafe fn run(&mut self, world: &World);
}

struct FuncSystem {
    sys: Box<dyn FnMut(&World)>,
}

impl System for FuncSystem {
    unsafe fn run(&mut self, world: &World) {
        (self.sys)(world);
    }
}

impl<P: SystemParam, F: FnMut(P) + FnMut(<P::Fetch as FetchSystemParam>::Item) + 'static>
    IntoSystem<P, FuncSystem> for F
{
    fn system(mut self) -> FuncSystem {
        FuncSystem {
            sys: Box::new(move |world: &World| {
                let param = P::Fetch::get_param(world);
                self(param);
            }),
        }
    }
}

impl<S: System> IntoSystem<(), S> for S {
    fn system(self) -> S {
        self
    }
}

struct World {
    value1: usize,
}

struct App {
    world: World,
    system: Option<Box<dyn System>>,
}

impl App {
    // only safe because system scope limited to self
    fn add_system<Params, S: System, IntoS: IntoSystem<Params, S>>(&mut self, into: IntoS) {
        let system = into.system();
        self.system = Some(Box::new(system));
    }

    fn run(&mut self) {
        // safe: world is uniquely accessed
        // fancy scheduler logic goes here
        unsafe {
            self.system.as_mut().unwrap().run(&self.world);
        }
    }
}

fn main() {
    fn sys(res: UsizeRes) {
        println!("{}", res.value);
    }

    // this cannot be converted to a system
    // fn sys_unsafe(res: UsizeRes<'static>) {
    //     println!("{}", res.value);
    // }

    let world = World { value1: 10 };
    // let mut system = sys.system();
    // system.run(&world);

    let mut app = App {
        world,
        system: None,
    };

    // this works
    app.add_system(sys.system());
    // this doesn't work
    // app.add_system(sys);
    app.run();
}
