/// This prevents UsizeRes<'static> by just removing lifetimes.
/// Its still horribly unsafe because UsizeRes can now be moved anywhere (creating dangling references)

use std::ops::Deref;

trait SystemParam {
    fn get_param(world: &World) -> Self;
}

struct UsizeRes {
    value: *const usize,
}

impl Deref for UsizeRes {
    type Target = usize;

    fn deref(&self) -> &Self::Target {
        unsafe { &*self.value }
    }
}

impl SystemParam for UsizeRes {
    fn get_param(world: &World) -> Self {
        UsizeRes {
            value: &world.value1,
        }
    }
}

trait IntoSystem<Params, Sys: System> {
    fn system(self) -> Sys;
}

trait System: 'static {
    fn run(&mut self, world: &World);
}

struct FuncSystem {
    sys: Box<dyn FnMut(&World)>,
}

impl System for FuncSystem {
    fn run(&mut self, world: &World) {
        (self.sys)(world);
    }
}

impl<P: SystemParam, F: FnMut(P) + 'static> IntoSystem<P, FuncSystem> for F {
    fn system(mut self) -> FuncSystem {
        FuncSystem {
            sys: Box::new(move |world: &World| {
                let param = P::get_param(world);
                self(param);
            }),
        }
    }
}

struct World {
    value1: usize,
}

struct App {
    world: World,
    system: Option<Box<dyn System>>,
}

impl App {
    // only safe because system scope limited to self
    fn add_system<'a, Params, S: System, IntoS: IntoSystem<Params, S>>(&mut self, into: IntoS) {
        let system = into.system();
        self.system = Some(Box::new(system));
    }

    fn run(&mut self) {
        self.system.as_mut().unwrap().run(&self.world);
    }
}

fn main() {
    fn sys(res: UsizeRes) {
        println!("{}", *res);
    }

    let world = World { value1: 10 };
    // let mut system = sys.system();
    // system.run(&world);

    let mut app = App {
        world,
        system: None,
    };

    app.add_system(sys);
    app.run();
}
