/// This is safe-ish. It relies on the app builder to ensure the lifetime passed in is tied to the relevant resources
/// Not ideal / significantly less safe than other options.
/// Additionally, because System has a lifetime, it is unbox-able 

trait SystemParam<'a> {
    fn get_param(world: &'a World) -> Self;
}

struct UsizeRes<'a> {
    value: &'a usize,
}

impl<'a> SystemParam<'a> for UsizeRes<'a> {
    fn get_param(world: &'a World) -> Self {
        UsizeRes {
            value: &world.value1,
        }
    }
}

trait IntoSystem<'a, Params, Sys: System<'a>> {
    unsafe fn system(self) -> Sys;
}

trait System<'a> {
    fn run(&mut self, world: &'a World);
}

struct FuncSystem<'a> {
    sys: Box<dyn FnMut(&'a World)>,
}

impl<'a> System<'a> for FuncSystem<'a> {
    fn run(&mut self, world: &'a World) {
        (self.sys)(world);
    }
}

impl<'a, P: SystemParam<'a>, F: FnMut(P) + 'static> IntoSystem<'a, P, FuncSystem<'a>> for F {
    unsafe fn system(mut self) -> FuncSystem<'a> {
        FuncSystem {
            sys: Box::new(move |world: &World| {
                let param = P::get_param(world);
                self(param);
            }),
        }
    }
}

struct World {
    value1: usize,
}

struct App {
    world: World,
}

impl App {
    // only safe because system scope limited to self
    fn run_system<'a, Params, S: System<'a>, IntoS: IntoSystem<'a, Params, S>>(
        &'a mut self,
        into: IntoS,
    ) {
        unsafe {
            let mut system = into.system();
            system.run(&self.world);
        }
    }
}

fn main() {
    fn sys(res: UsizeRes) {
        println!("{}", res.value);
    }

    let world = World { value1: 10 };
    // let mut system = sys.system();
    // system.run(&world);

    let mut app = App { world };

    app.run_system(sys);
}
